function varargout = mrstPath(varargin)
%Establish and maintain mapping from module names to system directory paths
%
% SYNOPSIS:
%   Either of the modes
%      1) mrstPath register list
%      2) mrstPath <command> [module list]
%      3) mrstPath search module list
%         paths = mrstPath('search', module list)
%
% PARAMETERS:
%   Mode 1)
%     Register (insert) new mappings of module names to system paths
%     (directories) into the current list.  In this case, the 'list' must
%     contain an even number of strings, the odd numbered ones being
%     interpreted as module names and the even numbered interpreted as
%     directories.  The directories must exist (i.e., function ISDIR must
%     return TRUE) to register a mapping.
%
%     EXAMPLE:
%        mrstPath register AGMG       modules/mex/AGMG       ...
%                          matlab_bgl modules/mex/matlab_bgl ...
%                          triangle   modules/mex/triangle
%
%   Mode 2)
%     <Command>  - One of the explicit verbs 'clear', 'list', 'remove'
%                  or 'reset'.  The semantics of the command verbs are as
%                  follows:
%
%                    o) clear  -- Deactivate all modules.  An explicit
%                                 module list, if present, is ignored.
%
%                    o) list   -- Display list of currently registered
%                                 modules in command window.  An explicit
%                                 module list, if present, is ignored.
%
%                    o) remove --
%
%                    o) reset  -- Convenience verb.  Equivalent to the verb
%                                 sequence:
%                                    mrstPath clear
%                                    mrstPath register [list]
%
%     [module list] - A sequence of strings naming individual add-on
%                     modules for MRST.  A module string/name may be either
%                     of the following:
%
%                       o) A relative path-name such as 'agglomeration' or
%                          'eclipse/resultinput'.  The name is interpreted
%                          as a directory relative to the default MRST
%                          module directory,
%
%                              fullfile(ROOTDIR, 'modules')
%
%                          If no such directory exists, as determined by
%                          ISDIR, the requested module is ignored.
%
%                          The string may include the path-name component
%                          separator, FILESEP.
%
%                       o) An absolute path-name identifying an arbitrary
%                          directory on the local computer system.  If the
%                          directory does not exist, the requested module
%                          is ignored.
%
%   Mode 2)
%     None.
%
% RETURNS:
%   Mode 1)
%     Nothing.
%
%   Mode 2)
%     modules - List, represented as a cell array of strings, of the
%               currently active add-on modules.
%
% EXAMPLES:
%    mrstModule add eclipse
%    mrstModule add /exp/module
%    mrstModule add ../utility/module
%
%    mrstModule list
%    mrstModule clear
%
% SEE ALSO:
%   ROOTDIR, isdir, filesep.

%{
Copyright 2009, 2010, 2011, 2012, 2013 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2013-10-18 18:00:12 +0200 (Fri, 18 Oct 2013) $
% $Revision: 11946 $

   persistent CACHE

   if isempty(CACHE), CACHE = struct(); end

   if nargin > 0,
      assert (iscellstr(varargin), 'All parameters must be strings.');

      cmd  = varargin{1};
      mods = varargin(2 : end);

      switch lower(cmd),
         case { 'add', 'register' },
            % Register a new set of modules.  Assume that 'mods' is a list
            % of key/value pairs.

            if ~ (isempty(mods) || all(cellfun(@isempty, mods))),
               mlock
               CACHE = register_modules(CACHE, mods);
            else
               munlock
            end

         case 'clear',
            % Clear cache, thus removing all currently registered modules
            % leaving only the modules that are known by default.

            munlock
            CACHE = clear_modules;

         case 'list',

            list_modules(CACHE);

         case 'remove',
            % Remove a module or set of modules.

            CACHE = remove_modules(CACHE, mods);

            if is_empty(CACHE),
               munlock
            end
            
         case 'reset',
            % Convenience verb: Clear existing list, then register new
            % modules.

            mrstPath clear

            if ~isempty(mods),
               [varargout{1:nargout}] = mrstPath('register', mods{:});
            end

         case { 'query', 'search' },
            % Look for a particular module or set of modules.  The 'mods'
            % list is interpreted as a list of modules to search for.

            [pth, notfound] = search_modules(CACHE, mods);

            if nargout > 0,
               lst = repmat({ '' }, [1, numel(mods)]);
               [lst{~notfound}] = pth{:};

               if (nargout == 1) && (numel(lst) > 1),

                  varargout{1} = lst;

               elseif nargout == numel(mods),

                  [varargout{1:numel(mods)}] = lst{:};

               end
               
            else

               map = [ reshape(mods(~notfound), [], 1) , ...
                       reshape(pth,             [], 1) ; ...
                       reshape(mods(notfound),  [], 1) , ...
                       repmat({ 'Not Found' }, [sum(notfound), 1]) ];

               fprintf('Module mapping results:\n');
               print_list(map);
            end

         otherwise

            error(msgid('Verb:Unknown'), ...
                 ['Unknown command verb ''%s''.  Must be one of ', ...
                  '''register'', ''clear'', ''list'', ''remove'', ', ...
                  '''reset'', or ''query''.'], cmd);

      end

   elseif nargout == 0,

      mrstPath list

   else

      error(msgid('Syntax:Error'), ...
           ['Call syntax is\n\t', ...
            mfilename, ' <command> [module list]  or\n\t', ...
            'map = ', mfilename]);

   end
end

%--------------------------------------------------------------------------

function cache = register_modules(cache, mods)
   assert (mod(numel(mods), 2) == 0, ...
           'Register: Must be list of key/value pairs');

   mods = reshape(mods, 2, []) .';
   i    = cellfun(@isdir, mods(:,2));

   if ~ all(i),
      if mrstVerbose,
         nonext = mods(~i, :);

         if size(nonext, 1) > 1,
            [pl1, pl2, pl3] = deal('s', 'ies', ''  );
         else
            [pl1, pl2, pl3] = deal('' , 'y'  , 'es');
         end

         fprintf('The following module%s could not be registered.\n', pl1);
         fprintf('Reason: The director%s do%s not exist.\n', pl2, pl3);

         print_list(nonext);
      end

      mods = mods(i, :);
   end

   map = cache_to_map(cache);
   [i, j] = look_for(map(:,1), mods);
   if ~isempty(i),
      map(i,2) = mods(j,2);
      mods     = mods(~j, :);
   end

   cache = map_to_cache([ map ; mods ]);
end

%--------------------------------------------------------------------------

function cache = clear_modules
   cache = struct();
end

%--------------------------------------------------------------------------

function tf = is_empty(cache)
   tf = isempty(cache_to_map(cache));
end

%--------------------------------------------------------------------------

function list_modules(cache)
   map = cache_to_map(cache);

   if isempty(map),
      fprintf('No modules registered\n');
   else
      fprintf('Currently registered modules\n');
      print_list(map);
   end
end

%--------------------------------------------------------------------------

function cache = remove_modules(cache, mods)
   map = cache_to_map(cache);
   i   = look_for(map(:,1), mods);

   pick    = false([size(map,1), 1]);
   pick(i) = true;

   cache = map_to_cache(map(~pick, :));
end

%--------------------------------------------------------------------------

function [pth, notfound] = search_modules(cache, mods)
   map    = cache_to_map(cache);
   [i, j] = look_for(map(:,1), mods);

   pth = map(i, 2);

   notfound     = false([numel(mods), 1]);
   notfound(~j) = true;
end

%--------------------------------------------------------------------------

function map = cache_to_map(cache)
   assert (numel(cache) == 1 && isstruct(cache));

   map = [ reshape(fieldnames(cache), [], 1) , struct2cell(cache) ];
end

%--------------------------------------------------------------------------

function cache = map_to_cache(map)
   assert (size(map, 2) == 2);

   cache = cell2struct(map(:,2), map(:,1), 1);
end

%--------------------------------------------------------------------------

function [match_i, match_j] = look_for(reg, mods)
   [i, j] = blockDiagIndex(numel(reg), numel(mods));

   matches = match(reshape(reg(i), [], 1), reshape(mods(j), [], 1));

   match_i = i(matches);
   match_j = false([numel(mods), 1]);
   match_j(j(matches)) = true;
end

%--------------------------------------------------------------------------

function tf = match(m1, m2)
   alg = match_algorithm;
   tf  = alg(m1, m2);
end

%--------------------------------------------------------------------------

function alg = match_algorithm
   if ispc,
      % Match module names irrespective of case on PC/WIN.
      alg = @strcmpi;
   else
      % Consider case when matching module names on Unix-like systems.
      alg = @strcmp;
   end
end

%--------------------------------------------------------------------------

function print_list(map)
   nchar = max(cellfun('prodofsize', map(:,1)));

   for k = 1 : size(map, 1),
      fprintf('  * %-*s -> %s\n', nchar, map{k,1}, map{k,2});
   end
end
